#!/usr/bin/perl -w

###
# Pod Documentation
###

=head1 NAME

cmsearch_.pl

=head1 SYNOPSIS

cmsearch_to_fasta.pl [-s 18.0 -n 10000] input_file.cms output_file.fna

Convert Infernal cmsearch results to a FASTA file sorted by bit score, 
that is suitable for input into cmalign.

=head1 DESCRIPTION

=over

=item B<-t or -T> <bit score> 

Only collect hits above this score threshhold.

=item B<-n> <unsigned integer> 

End after this maximum number of hits has been collected starting with the highest scoring hits.

=item B<-s or -b> <bit score file name> 

Create a tab-delimited file of bit scores for each sequence. (For use as input to genomic_context.pl)

=back

=head1 AUTHOR

Jeffrey Barrick

=head1 COPYRIGHT

Copyright (C) 2008-2009 Jeffrey Barrick.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=cut

###
# End Pod Documentation
###

use strict;

use FindBin;
use lib $FindBin::Bin;
use Bio::DB::Flat;
use Bio::SearchIO::infernal;
use Bio::SeqIO;
use Getopt::Long;
use Pod::Usage;

##Handle command line options
my ($help, $man);
my $score_threshold;
my $number_threshold;
my $score_file;
my $score_file_2;
our ($db_path, $db_dir, $db_name);
our $verbose = 0;
pod2usage(1) if (scalar @ARGV == 0);
GetOptions(
	'help|?' => \$help, 'man' => \$man,
	'score-threshold|t=s' => \$score_threshold,
	'number-threshold|n=s' => \$number_threshold,
	'score-file|s=s' => \$score_file,
	'bit-score-file|b=s' => \$score_file_2,
	'database-name|d=s' => \$db_path,
) or pod2usage(2);
$score_file = $score_file_2 if ( (defined $score_file_2) && (!defined $score_file));

if ($db_path)
{
	$db_path =~ s/\/$//;
	$db_path =~ m/^(.+)\/(.+)$/;
	($db_dir, $db_name) = ($1, $2);
}

my ($input_file, $output_file) = @ARGV;
pod2usage(1) if $help;
pod2usage(-exitstatus => 0, -verbose => 2) if $man;
pod2usage(1) if (!defined $input_file or !defined $output_file);

##Open output files
my $out = Bio::SeqIO->new(-file => ">$output_file", -format => 'fasta');
(open SCORES, ">$score_file" or die "Could not open scores file $score_file" ) if ($score_file);

##Read in hits
my $searchio = Bio::SearchIO::infernal->new(-file => "<$input_file");

use Data::Dumper;
my @hsps;
my $num_hits = 0;
my $loaded_hits = 0;
RESULT: while (my $result = $searchio->next_result)
{	
	while(my $hit = $result->next_hit) 
	{	
		while(my $hsp = $hit->next_hsp) 
		{
			$loaded_hits++;
			print STDERR "Loaded $loaded_hits hits...\n" if ($loaded_hits % 100 == 0);
			push @hsps, $hsp if ((!defined $score_threshold) || ($hsp->bits >= $score_threshold));
			$num_hits++;
		
			#uncomment for debugging
			#last RESULT if ($num_hits > 1000);
			#print Dumper($hsp);
		}
	}
}

print "Loaded $num_hits hits from \"$input_file\".\n";
$number_threshold = scalar @hsps if ((!defined $number_threshold) || ($number_threshold > scalar @hsps));

#Create a new list of sequences from hsps so that we can sort and reformat sequences.
my @hit_seqs;
foreach my $hsp (@hsps)
{
	my $seq = $hsp->hit_string;
	my $feat = $hsp->hit;
	my $disp_name = $feat->seq_id();

	#For RefSeq and GenBank records, set the name to the accession number
	$disp_name =~ m/ref\|(.+?)[ $|]/;
	if ($1)
	{
		#die "Could not find accession number of record: $disp_name" if (!$1);
		$disp_name = $1;
	}	
	
	#Use start-end coordinates in the name
	my ($start, $end) = ($feat->start, $feat->end);
	($start, $end) = ($end, $start) if ($feat->strand == -1); 
	$disp_name .= "/$start-$end";
		
	push @hit_seqs, { 
		seq => $seq, 
		display_id => $disp_name, 
		bits => $hsp->bits,
		start => $feat->start,
		end => $feat->end,
		strand => $feat->strand,
	};
}


##Sort by score #Secondarily by sequence accession number
sub by_score_accession
{	
	my $a_name = $a->{display_id};
	#Special RefSeq sorting rules
	if ($a_name =~ m/^NC_/)
	{
		$a_name = "01$a_name";
	}
	else
	{
		$a_name = "02$a_name";
	}
	
	my $b_name = $b->{display_id};
	if ($b_name =~ m/^NC_/)
	{
		$b_name = "01$b_name";
	}
	else
	{
		$b_name = "02$b_name";
	}
	return ( -($a->{bits} <=> $b->{bits}) || ($a_name cmp $b_name) );
}

@hit_seqs = sort by_score_accession @hit_seqs;

##Output
for (my $i=0; $i < $number_threshold; $i++)
{	
	my $seq = $hit_seqs[$i];
	
	#if the search was run in local mode then there may be bracketed numbers
	#of bases that were not output, we need to grab the sequence in this case
	if ($seq->{seq} =~ m/(\[|\]|\*)/)
	{	
		my $accession = $seq->{display_id};
		$accession =~ s/\/.+$//;
		my $reference_seq = load_fasta_from_database($accession);
		my $temp_seq = $reference_seq->trunc($seq->{start}, $seq->{end});
		$temp_seq = $temp_seq->revcom() if (defined $seq->{strand} && ($seq->{strand} == -1));
		$seq->{seq} = $temp_seq->seq;
	}
	else #can directly use loaded sequence
	{
		$seq->{seq} = "\U$seq->{seq}";
		$seq->{seq} =~ s/[^ATCGU]//g;	
	}
	
	my $s = new Bio::Seq(-display_id => $seq->{display_id},-seq => $seq->{seq} );
	
	$out->write_seq($s);
	
	if ($score_file)
	{
		print SCORES $seq->{display_id} . "\t" . $seq->{bits} . "\n";
	}
}

print "Wrote $number_threshold hits to \"$output_file\".\n";
print "Wrote scores to \"$score_file\".\n" if ($score_file);


sub load_fasta_from_database
{
	my ($fragment_accession) = @_;
	die "No FASTA database path provided" if (!defined $db_dir);
	
	print STDERR "Loading $fragment_accession from database.\n" if ($verbose);
	my $database = Bio::DB::Flat->new(	-directory  => $db_dir,
                    -dbname     => $db_name,
                    -format     => 'fasta',
                    -index      => 'flat',
                    -write_flag => 0);
	my $seq = $database->get_Seq_by_id($fragment_accession);
	
	die "Could not find sequence with accession $fragment_accession in database\n" if (!$seq);

	return $seq;
}
